/*
 * *************************************************
 * Copyright (c) 2018 QMTH. All Rights Reserved.
 * Created by Deason on 2018-10-22 15:27:49.
 * *************************************************
 */

package cn.com.qmth.examcloud.core.print.service.impl;

import cn.com.qmth.examcloud.core.print.common.jpa.OrderBuilder;
import cn.com.qmth.examcloud.core.print.common.jpa.SearchBuilder;
import cn.com.qmth.examcloud.core.print.common.jpa.SpecUtils;
import cn.com.qmth.examcloud.core.print.common.jpa.SqlWrapper;
import cn.com.qmth.examcloud.core.print.common.utils.Check;
import cn.com.qmth.examcloud.core.print.entity.CourseStatistic;
import cn.com.qmth.examcloud.core.print.enums.PaperStatus;
import cn.com.qmth.examcloud.core.print.repository.CourseStatisticRepository;
import cn.com.qmth.examcloud.core.print.service.CourseStatisticService;
import cn.com.qmth.examcloud.core.print.service.PrintingProjectService;
import cn.com.qmth.examcloud.core.print.service.StatisticService;
import cn.com.qmth.examcloud.core.print.service.bean.common.CourseInfo;
import cn.com.qmth.examcloud.core.print.service.bean.common.ExamCourseInfo;
import cn.com.qmth.examcloud.core.print.service.bean.common.ExamInfo;
import cn.com.qmth.examcloud.core.print.service.bean.coursestatistic.*;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.jdbc.core.BeanPropertyRowMapper;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * @author: fengdesheng
 * @since: 2018/10/17
 */
@Service
public class CourseStatisticServiceImpl implements CourseStatisticService {
    private static final Logger log = LoggerFactory.getLogger(CourseStatisticServiceImpl.class);
    @Autowired
    private CourseStatisticRepository courseStatisticRepository;
    @Autowired
    private PrintingProjectService printingProjectService;
    @Autowired
    private StatisticService statisticService;
    @Autowired
    private JdbcTemplate jdbcTemplate;

    @Override
    public Page<CourseStatisticInfo> getCourseStatisticList(CourseStatisticQuery query) {
        Check.isNull(query, "查询参数不能为空！");

        //查询条件
        SearchBuilder searches = new SearchBuilder();
        if (query.getOrgId() != null) {
            searches.eq("orgId", query.getOrgId());
        }
        if (query.getExamId() != null) {
            searches.eq("examId", query.getExamId());
        }
        if (query.getCourseId() != null) {
            searches.eq("courseId", query.getCourseId());
        }
        if (query.getPaperStatus() != null) {
            searches.eq("paperStatus", query.getPaperStatus());
        }
        if (StringUtils.isNotBlank(query.getPaperName())) {
            searches.like("coursePaper.paperName", query.getPaperName());
        }
        Specification<CourseStatistic> spec = SpecUtils.buildSearchers(CourseStatistic.class, searches.build());
        //排序条件
        OrderBuilder orders = new OrderBuilder().desc("id");

        //分页条件
        Pageable pageable = SpecUtils.buildPageable(query.getPageNo(), query.getPageSize(), orders.build());

        Page<CourseStatistic> page = courseStatisticRepository.findAll(spec, pageable);
        return CourseStatisticConvert.ofPage(page);
    }

    @Override
    public void syncCourseNameByCourseId(Long courseId, String courseName) {
        if (courseId == null || StringUtils.isBlank(courseName)) {
            return;
        }
        courseStatisticRepository.updateCourseNameByCourseId(courseId, courseName);
    }

    @Override
    public List<CourseStatisticLessInfo> getCourseStatisticLessInfoList(Long orgId, Long examId) {
        Check.isNull(orgId, "学校ID不能为空！");
        Check.isNull(examId, "考试ID不能为空！");

        SqlWrapper sql = new SqlWrapper()
                .select("cs.id,cs.total_student,cp.paper_p").from("ec_prt_course_statistic cs")
                .innerJoin("ec_prt_course_paper cp").on("cp.id", "cs.course_paper_id")
                .where().eq("cs.org_id", orgId).and().eq("cs.exam_id", examId);
        return jdbcTemplate.query(sql.build(), new BeanPropertyRowMapper(CourseStatisticLessInfo.class));
    }

    @Override
    public void refreshCourseStatistic(CourseStatisticRefreshReq req) {
        Check.isNull(req, "请求参数不能为空！");
        Check.isNull(req.getOrgId(), "学校ID不能为空！");
        Check.isNull(req.getExamId(), "考试ID不能为空！");

        List<CourseStatisticRefreshReq.Course> courses = req.getCourses();
        if (courses == null || courses.size() == 0) {
            //未选择课程时，则跳过统计
            return;
        }

        //获取当前考试下已有的课程统计信息
        Map<String, CourseStatistic> oldCourseStatisticMaps = this.getCourseStatisticMaps(req.getOrgId(), req.getExamId());

        //待刷新的课程列表
        List<ExamCourseInfo> examCourses = new ArrayList<>();
        for (CourseStatisticRefreshReq.Course course : courses) {
            if (course.getCourseId() == null || StringUtils.isBlank(course.getPaperType())) {
                continue;
            }
            examCourses.add(new ExamCourseInfo(req.getOrgId(), req.getExamId(), course.getCourseId(), course.getPaperType()));
        }

        this.syncCourseStatistics(examCourses, oldCourseStatisticMaps);
    }

    @Override
    public void initAllCourseStatistic() {
        log.debug("initAllCourseStatistic...");

        //只处理印刷项目中已有的考试
        List<ExamInfo> exams = printingProjectService.getExamList(null);
        if (exams == null || exams.isEmpty()) {
            return;
        }

        for (ExamInfo exam : exams) {
            //获取当前考试下已有的课程统计信息
            Map<String, CourseStatistic> oldCourseStatisticMaps = this.getCourseStatisticMaps(exam.getOrgId(), exam.getExamId());

            //获取某考试的课程和试卷类型列表
            List<ExamCourseInfo> examCourses = statisticService.findExamCourseAndPaperTypes(exam.getOrgId(), exam.getExamId());

            this.syncCourseStatistics(examCourses, oldCourseStatisticMaps);
        }
    }

    private void syncCourseStatistics(List<ExamCourseInfo> examCourses, Map<String, CourseStatistic> oldCourseStatisticMaps) {
        if (examCourses == null || examCourses.isEmpty()) {
            return;
        }

        //课程基本信息Maps
        Map<Long, CourseInfo> courseMaps = new HashMap<>();

        //按课程和试卷类型逐个更新统计信息
        for (ExamCourseInfo info : examCourses) {
            //Key: examId_courseId_paperType
            String key = info.getExamId() + "_" + info.getCourseId() + "_" + info.getPaperType();

            //获取当前考生的数量
            int totalStudent = statisticService.countExamTotalStudentByCourseIdAndPaperType(info.getExamId(), info.getCourseId(), info.getPaperType());
            info.setTotalStudent(totalStudent);

            if (oldCourseStatisticMaps.containsKey(key)) {
                //已存在则更新
                this.updateTotalStudentByOrgExamCourse(info);
                oldCourseStatisticMaps.remove(key);
            } else {
                //获取并设置课程基本信息
                CourseInfo course = courseMaps.get(info.getCourseId());
                if (course == null) {
                    course = statisticService.findCourseInfo(info.getOrgId(), info.getCourseId());
                    courseMaps.put(course.getId(), course);
                }
                info.setCourseCode(course.getCode());
                info.setCourseName(course.getName());

                //不存在则新增
                this.addCourseStatistic(info);
            }
        }

        //清除已不存在的考试课程
        if (oldCourseStatisticMaps.size() > 0) {
            courseStatisticRepository.delete(oldCourseStatisticMaps.values());
        }
    }

    private void addCourseStatistic(ExamCourseInfo info) {
        CourseStatistic statistic = new CourseStatistic();
        statistic.setOrgId(info.getOrgId());
        statistic.setExamId(info.getExamId());
        statistic.setCourseId(info.getCourseId());
        statistic.setCourseCode(info.getCourseCode());
        statistic.setCourseName(info.getCourseName());
        statistic.setPaperType(info.getPaperType());
        statistic.setTotalStudent(info.getTotalStudent());
        statistic.setPaperStatus(PaperStatus.无.getIndex());
        courseStatisticRepository.save(statistic);
    }

    private void updateTotalStudentByOrgExamCourse(ExamCourseInfo info) {
        if (info == null || info.getTotalStudent() == null) {
            return;
        }
        //仅更新total_student字段
        courseStatisticRepository.updateTotalStudentByOrgExamCourse(info.getOrgId(), info.getExamId(), info.getCourseId(), info.getPaperType(), info.getTotalStudent());
    }

    private Map<String, CourseStatistic> getCourseStatisticMaps(Long orgId, Long examId) {
        SearchBuilder searches = new SearchBuilder()
                .eq("orgId", orgId)
                .eq("examId", examId);
        Specification<CourseStatistic> spec = SpecUtils.buildSearchers(CourseStatistic.class, searches.build());
        List<CourseStatistic> list = courseStatisticRepository.findAll(spec);

        Map<String, CourseStatistic> maps = new HashMap<>();
        if (list != null && !list.isEmpty()) {
            for (CourseStatistic statistic : list) {
                //Key: examId_courseId_paperType
                String key = statistic.getExamId() + "_" + statistic.getCourseId() + "_" + statistic.getPaperType();
                maps.put(key, statistic);
            }
        }
        return maps;
    }

}